// vim:filetype=cpp:textwidth=120:shiftwidth=2:softtabstop=2:expandtab
// Copyright 2016 Christoph Schwering
// Licensed under the MIT license. See LICENSE file in the project root.
//
// Traits, mostly to extract and identify the arguments of functors.

#ifndef LIMBO_INTERNAL_TRAITS_H_
#define LIMBO_INTERNAL_TRAITS_H_

#include <tuple>
#include <type_traits>

namespace limbo {
namespace internal {

// first_type<T, U>::type is T::type if it exists, and U::type else.
template<typename>
struct first_type_helper {
  typedef void type;
};

template<typename T, typename U, typename = void>
struct first_type {
  typedef typename U::type type;
};

template<typename T, typename U>
struct first_type<T, U, typename first_type_helper<typename T::type>::type> {
  typedef typename T::type type;
};

// arg<Function>::type<N> is the type of the Nth argument of Function().
template<typename Function>
struct arg : public arg<decltype(&Function::operator())> {};

template<typename Function, typename Return, typename... Args>
struct arg<Return (Function::*)(Args...) const> {
  template<int N = 0>
  using type = typename std::tuple_element<N, typename std::tuple<Args...>>::type;
};

// is_arg<Function, ExpType, N>::value is true iff ExpType is convertible to
// Nth argument of Function, and false otherwise.
template<typename Function, typename ExpType, int N = 0>
struct is_arg {
  static constexpr bool value = std::is_convertible<ExpType, typename arg<Function>::template type<N>>::value;
};

// if_arg<Function, ExpType, Type, N>::type is defined as Type iff ExpType
// is convertible to the Nth argument of Function.
template<typename Function, typename ExpType, typename Type = void, int N = 0,
  bool = is_arg<Function, ExpType, N>::value>
struct if_arg {};

template<typename Function, typename ExpType, typename Type, int N>
struct if_arg<Function, ExpType, Type, N, true> {
  typedef Type type;
};

// remove_const_ref<T>::type removes the const and reference from T.
template<typename T>
struct remove_const_ref {
  typedef typename std::remove_const<typename std::remove_reference<T>::type>::type type;
};

}  // namespace internal
}  // namespace limbo

#endif  // LIMBO_INTERNAL_TRAITS_H_


